@@ -2,24 +2,26 @@
 
 # geoarray, A fast Python interface for image geodata - either on disk or in memory.
 #
-# Copyright (C) 2019  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
+# Copyright (C) 2017-2021
+# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
+# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
+#   Germany (https://www.gfz-potsdam.de/)
 #
 # This software was developed within the context of the GeoMultiSens project funded
 # by the German Federal Ministry of Education and Research
 # (project grant code: 01 IS 14 010 A-C).
 #
-# This program is free software: you can redistribute it and/or modify it under
-# the terms of the GNU Lesser General Public License as published by the Free
-# Software Foundation, either version 3 of the License, or (at your option) any
-# later version.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
-# details.
+#   http://www.apache.org/licenses/LICENSE-2.0
 #
-# You should have received a copy of the GNU Lesser General Public License along
-# with this program.  If not, see <http://www.gnu.org/licenses/>.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 
 import os
@@ -27,10 +29,10 @@ import warnings
 from pkgutil import find_loader
 from collections import OrderedDict
 from copy import copy, deepcopy
-from typing import Union  # noqa F401
+from typing import Union, Optional  # noqa F401
 
 import numpy as np
-from osgeo import gdal, gdal_array, gdalnumeric
+from osgeo import gdal, gdal_array
 from shapely.geometry import Polygon
 from shapely.wkt import loads as shply_loads
 # dill -> imported when dumping GeoArray
@@ -48,7 +50,7 @@ from py_tools_ds.io.raster.gdal import get_GDAL_ds_inmem
 from py_tools_ds.numeric.numbers import is_number
 from py_tools_ds.numeric.array import get_array_tilebounds
 
-#  internal imports
+# internal imports
 from .subsetting import get_array_at_mapPos
 from .metadata import GDAL_Metadata
 
@@ -64,9 +66,9 @@ class GeoArray(object):
     some functions for quickly visualizing the data as a map, a simple image or an interactive image.
     """
 
-    def __init__(self, path_or_array, geotransform=None, projection=None, bandnames=None, nodata=None, progress=True,
-                 q=False):
-        # type: (Union[str, np.ndarray, GeoArray], tuple, str, list, float, bool, bool) -> None
+    def __init__(self, path_or_array, geotransform=None, projection=None, bandnames=None, nodata=None, basename='',
+                 progress=True, q=False):
+        # type: (Union[str, np.ndarray, GeoArray], tuple, str, list, float, str, bool, bool) -> None
         """Get an instance of GeoArray.
 
         :param path_or_array:   a numpy.ndarray or a valid file path
@@ -76,6 +78,7 @@ class GeoArray(object):
         :param bandnames:       names of the bands within the input array, e.g. ['mask_1bit', 'mask_clouds'],
                                 (default: ['B1', 'B2', 'B3', ...])
         :param nodata:          nodata value
+        :param basename:        a short base name of the dataset (e.g., used in some status messages)
         :param progress:        show progress bars (default: True)
         :param q:               quiet mode (default: False)
         """
@@ -90,7 +93,9 @@ class GeoArray(object):
         if isinstance(path_or_array, str):
             assert ' ' not in path_or_array, "The given path contains whitespaces. This is not supported by GDAL."
 
-            if not os.path.exists(path_or_array):
+            if not os.path.exists(path_or_array) and \
+               not path_or_array.startswith('/vsi') and \
+               not path_or_array.startswith('HDF'):
                 raise FileNotFoundError(path_or_array)
 
         if isinstance(path_or_array, GeoArray) or issubclass(getattr(path_or_array, '__class__'), GeoArray):
@@ -99,16 +104,18 @@ class GeoArray(object):
             self.geotransform = geotransform or self.geotransform
             self.projection = projection or self.projection
             self.bandnames = bandnames or list(self.bandnames.keys())
+            self.basename = basename or self.basename
             self._nodata = nodata if nodata is not None else self._nodata
             self.progress = False if progress is False else self.progress
-            self.q = q if q is not None else self.q
+            self.q = q or self.q
 
         else:
             self._initParams = dict([x for x in locals().items() if x[0] != "self"])
             self.arg = path_or_array
             self._arr = path_or_array if isinstance(path_or_array, np.ndarray) else None
             self.filePath = path_or_array if isinstance(path_or_array, str) and path_or_array else None
-            self.basename = os.path.splitext(os.path.basename(self.filePath))[0] if not self.is_inmem else 'IN_MEM'
+            basename_default = os.path.splitext(os.path.basename(self.filePath))[0] if not self.is_inmem else 'IN_MEM'
+            self.basename = basename or basename_default
             self.progress = progress
             self.q = q
             self._arr_cache = None  # dict containing key 'pos' and 'arr_cached'
@@ -149,8 +156,11 @@ class GeoArray(object):
             return
 
         if not dt_min <= _nodata <= dt_max:
-            raise ValueError("The given no-data value (%s) is out range for data type %s."
-                             % (self._initParams['nodata'], str(np.dtype(self.dtype))))
+            if np.issubdtype(self.dtype, np.floating) and np.isnan(_nodata):
+                pass
+            else:
+                raise ValueError("The given no-data value (%s) is out range for data type %s."
+                                 % (self._initParams['nodata'], str(np.dtype(self.dtype))))
 
     @property
     def arr(self):
@@ -168,6 +178,8 @@ class GeoArray(object):
             self.flush_cache()  # the cached array is not useful anymore
 
         self._arr = ndarray
+        self._dtype = ndarray.dtype
+        self._shape = ndarray.shape
 
     @property
     def bandnames(self):
@@ -187,8 +199,11 @@ class GeoArray(object):
                                 "Received %s." % type(list_bandnames))
             if len(list_bandnames) != self.bands:
                 raise ValueError('Number of given bandnames does not match number of bands in array.')
-            if len(list(set([type(b) for b in list_bandnames]))) != 1 or not isinstance(list_bandnames[0], str):
-                raise ValueError("'bandnames must be a set of strings. Got other datatypes in there.'")
+            if len(list(set([type(b) for b in list_bandnames]))) != 1:
+                raise ValueError('Multiple data types of the band names are not supported.')
+            if not isinstance(list_bandnames[0], (str, int, float)):
+                raise ValueError(f'Band names must be a set of strings, integers, or floats. '
+                                 f'Got {type(list_bandnames[0])}')
 
             bN_dict = OrderedDict((band, i) for i, band in enumerate(list_bandnames))
 
@@ -253,13 +268,14 @@ class GeoArray(object):
     @property
     def dtype(self):
         """Get the numpy data type of the associated image array."""
-        if self._dtype:
-            return self._dtype
-        elif self.is_inmem:
+        if self.is_inmem:
             return self.arr.dtype
         else:
-            self.set_gdalDataset_meta()
-            return self._dtype
+            if self._dtype:
+                return self._dtype
+            else:
+                self.set_gdalDataset_meta()
+                return self._dtype
 
     @property
     def geotransform(self):
@@ -349,12 +365,18 @@ class GeoArray(object):
         """Return 'True' if the image has a valid geoinformation with map instead of image coordinates."""
         return all([self.gt, list(self.gt) != [0, 1, 0, 0, 0, -1], self.prj])
 
+    @property
+    def is_rotated(self):
+        # type: () -> bool
+        """Return 'True' if the image has a rotation in the map info (i.e., is pseudo-projected)."""
+        return self.gt[2] != 0 or self.gt[4] != 0
+
     @property
     def nodata(self):
         """Get the nodata value of the GeoArray instance.
 
         If GeoArray has been instanced with a file path the metadata of the file on disk is checked for an existing
-        nodata value. Otherwise (if no value is exlicitly given during object instanciation) an automatic detection
+        nodata value. Otherwise, (if no value is exlicitly given during object instanciation) an automatic detection
         based on 3x3 windows at each image corner is run that analyzes the mean and standard deviation of these windows.
         """
         if self._nodata is not None:
@@ -376,7 +398,14 @@ class GeoArray(object):
 
     @nodata.setter
     def nodata(self, value):
-        # type: (Union[int, None]) -> None
+        # type: (Optional[bool, int, float]) -> None
+        if isinstance(value, np.bool_):
+            value = bool(value)
+        elif isinstance(value, np.integer):
+            value = int(value)
+        elif isinstance(value, np.floating):
+            value = float(value)
+
         self._nodata = value
 
         if self._metadata and value is not None:
@@ -475,7 +504,7 @@ class GeoArray(object):
             else:
                 try:
                     multipolygon = raster2polygon(self.mask_nodata.astype(np.uint8), self.gt, self.prj, exact=False,
-                                                  progress=self.progress, q=self.q, maxfeatCount=10, timeout=15)
+                                                  min_npx=10, progress=self.progress, q=self.q, timeout=15)
                     self._footprint_poly = fill_holes_within_poly(multipolygon)
                 except (RuntimeError, TimeoutError):
                     if not self.q:
@@ -507,13 +536,14 @@ class GeoArray(object):
 
     @property
     def metadata(self):
+        # type: () -> GDAL_Metadata
         """Return a DataFrame containing all available metadata (read from file if available).
 
         Use 'metadata[band_index].to_dict()' to get a metadata dictionary for a specific band.
         Use 'metadata.loc[row_name].to_dict()' to get all metadata values of the same key for all bands as dictionary.
         Use 'metadata.loc[row_name, band_index] = value' to set a new value.
 
-        :return:  pandas.DataFrame
+        :return:    instance of GDAL_Metadata
         """
         if self._metadata is not None:
             return self._metadata
@@ -535,7 +565,7 @@ class GeoArray(object):
                              % (self.__class__.__name__, self.__class__.__name__))
         self._metadata = meta
 
-    meta = alias_property('metadata')
+    meta = alias_property('metadata')  # type: GDAL_Metadata
 
     def __getitem__(self, given):
         if isinstance(given, (int, float, slice, np.integer, np.floating)) and self.ndim == 3:
@@ -935,7 +965,7 @@ class GeoArray(object):
             # TODO insert a multiprocessing.Lock here in order to prevent IO bottlenecks?
             # read subset area from disk
             if bL == list(range(0, B)):
-                tempArr = gdalnumeric.LoadFile(path, cS, rS, cE - cS + 1, rE - rS + 1)
+                tempArr = gdal_array.LoadFile(path, cS, rS, cE - cS + 1, rE - rS + 1)
                 out_arr = np.swapaxes(np.swapaxes(tempArr, 0, 2), 0, 1) if B > 1 else tempArr
                 if out_arr is None:
                     raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
@@ -987,91 +1017,72 @@ class GeoArray(object):
 
         driver = gdal.GetDriverByName(fmt)
         if driver is None:
-            raise Exception("'%s' is not a supported GDAL driver. Refer to www.gdal.org/formats_list.html for full "
-                            "list of GDAL driver codes." % fmt)
+            raise Exception("'%s' is not a supported GDAL driver. Refer to https://gdal.org/drivers/raster/index.html "
+                            "for full list of GDAL driver codes." % fmt)
 
         if not os.path.isdir(os.path.dirname(out_path)):
             os.makedirs(os.path.dirname(out_path))
 
         envi_metadict = self.metadata.to_ENVI_metadict()
 
-        #####################
-        # write raster data #
-        #####################
+        ###########################
+        # get source GDAL dataset #
+        ###########################
+
+        ds_src: gdal.Dataset
+        ds_out: gdal.Dataset
 
         if self.is_inmem:
-            # write in-memory data
-            ds_inmem = get_GDAL_ds_inmem(self.arr,  # expects rows,columns,bands
-                                         self.geotransform, self.projection,
-                                         self._nodata)  # avoid to compute the nodata value here, so use private attrib.
-
-            # write dataset
-            ds_out = driver.CreateCopy(out_path, ds_inmem,
-                                       options=creationOptions or [])  # noqa
-
-            # # rows, columns, bands => bands, rows, columns
-            # out_arr = self.arr if self.ndim == 2 else np.swapaxes(np.swapaxes(self.arr, 0, 2), 1, 2)
-            # gdalnumeric.SaveArray(out_arr, out_path, format=fmt, prototype=ds_inmem)  # expects bands,rows,columns
-            # ds_out = gdal.Open(out_path)
-
-            del ds_inmem
-            del ds_out
+            ds_src = get_GDAL_ds_inmem(self.arr,  # expects rows,columns,bands
+                                       self.geotransform, self.projection,
+                                       self._nodata)  # avoid to compute the nodata value here, so use private attrib.
 
         else:
-            # transform linked image dataset (not in-memory) to target format
-            src_ds = gdal.Open(self.filePath)
-            if not src_ds:
-                raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
-
+            ds_src = gdal.Open(self.filePath)
             # metadomains = {dom: src_ds.GetMetadata(dom) for dom in src_ds.GetMetadataDomainList()}
 
-            gdal.Translate(out_path, src_ds, format=fmt, creationOptions=creationOptions)
-            del src_ds
+        if not ds_src:
+            raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
 
-        if not os.path.exists(out_path):
-            raise Exception(gdal.GetLastErrorMsg())
-
-        ################
-        # set metadata #
-        ################
-
-        # NOTE:  The dataset has to be written BEFORE metadata are added. Otherwise, metadata are not written.
-
-        ds_out = gdal.Open(out_path, gdal.GA_Update)
+        #########################################
+        # write output dataset and set metadata #
+        #########################################
 
         try:
-            if not ds_out:
-                raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
-
             # ENVI #
             ########
             if fmt == 'ENVI':
+                # NOTE:  The dataset has to be written BEFORE metadata are added. Otherwise, metadata are not written.
+
+                # write ds_src to disk and re-open it to add the metadata
+                gdal.Translate(out_path, ds_src, format=fmt, creationOptions=creationOptions)
+                del ds_src
+                ds_out = gdal.Open(out_path, gdal.GA_Update)
+
                 for bidx in range(self.bands):
                     band = ds_out.GetRasterBand(bidx + 1)
 
                     if 'band_names' in envi_metadict:
-                        bandname = self.metadata.band_meta['band_names'][bidx].strip()
+                        bandname = str(self.metadata.band_meta['band_names'][bidx]).strip()
                         band.SetDescription(bandname)
                         assert band.GetDescription() == bandname
 
-                    # NOTE: band-specific nodata values are not supported by the ENVI header format
-                    if 'nodata' in envi_metadict:
-                        nodataVal = self.metadata.band_meta['nodata'][bidx]
-                        band.SetNoDataValue(nodataVal)
-                        assert band.GetNoDataValue() == nodataVal
-
                     del band
 
                 # avoid that band names are written to global meta
                 if 'band_names' in envi_metadict:
                     del envi_metadict['band_names']
 
+                # the expected key name is 'data_ignore_value', see below
                 if 'nodata' in envi_metadict:
                     del envi_metadict['nodata']
 
                 # set data_ignore_value in case self.metadata.band_meta contains a unique nodata value
-                if 'nodata' in self.metadata.band_meta and len(set(self.metadata.band_meta['nodata'])) == 1:
-                    envi_metadict['data_ignore_value'] = str(self.metadata.band_meta['nodata'][0])
+                if 'nodata' in self.metadata.band_meta:
+                    if len(set(self.metadata.band_meta['nodata'])) == 1:
+                        envi_metadict['data_ignore_value'] = str(self.metadata.band_meta['nodata'][0])
+                    else:
+                        warnings.warn("Band-specific nodata values are not supported by the ENVI header format.")
 
                 ds_out.SetMetadata(envi_metadict, 'ENVI')
 
@@ -1079,44 +1090,66 @@ class GeoArray(object):
                     ds_out.SetDescription(envi_metadict['description'])
 
                 ds_out.FlushCache()
+
+                # NOTE: In case of ENVI format and GDAL_PAM_ENABLED=NO, the metadata is not written
                 gdal.Unlink(out_path + '.aux.xml')
 
-            elif self.metadata.all_meta:
-                # set global domain metadata
-                if self.metadata.global_meta:
-                    ds_out.SetMetadata(dict((k, repr(v)) for k, v in self.metadata.global_meta.items()))
+            else:
+                # disable to write separate metadata XML files
+                os.environ['GDAL_PAM_ENABLED'] = 'NO'
 
-                if 'description' in envi_metadict:
-                    ds_out.SetDescription(envi_metadict['description'])
+                ds_out = ds_src
 
-                # set band domain metadata
-                bandmeta_dict = self.metadata.to_DataFrame().astype(str).to_dict()
+                # set metadata
+                if self.metadata.all_meta:
 
-                for bidx in range(self.bands):
-                    band = ds_out.GetRasterBand(bidx + 1)
-                    bandmeta = bandmeta_dict[bidx].copy()
+                    # set global domain metadata
+                    if self.metadata.global_meta:
+                        ds_out.SetMetadata(dict((k, repr(v)) for k, v in self.metadata.global_meta.items()))
 
-                    # filter global metadata out
-                    bandmeta = {k: v for k, v in bandmeta.items() if k not in self.metadata.global_meta}
-                    # meta2write = dict((k, repr(v)) for k, v in self.metadata.band_meta.items() if v is not np.nan)
+                    if 'description' in envi_metadict:
+                        ds_out.SetDescription(envi_metadict['description'])
 
-                    if 'band_names' in bandmeta:
-                        band.SetDescription(self.metadata.band_meta['band_names'][bidx].strip())
-                        del bandmeta['band_names']
+                    # set band domain metadata
+                    bandmeta_dict = self.metadata.to_DataFrame().astype(str).to_dict()
 
-                    if 'nodata' in bandmeta:
-                        band.SetNoDataValue(self.metadata.band_meta['nodata'][bidx])
-                        del bandmeta['nodata']
+                    for bidx in range(self.bands):
+                        band = ds_out.GetRasterBand(bidx + 1)
+                        bandmeta = bandmeta_dict[bidx].copy()
 
-                    if bandmeta:
-                        band.SetMetadata(bandmeta)
+                        # filter global metadata out
+                        bandmeta = {k: v for k, v in bandmeta.items() if k not in self.metadata.global_meta}
+                        # meta2write = dict((k, repr(v)) for k, v in self.metadata.band_meta.items() if v is not np.nan)
 
-                    band.FlushCache()
-                    del band
+                        if 'band_names' in bandmeta:
+                            bandname = str(self.metadata.band_meta['band_names'][bidx]).strip()
+                            band.SetDescription(bandname)
+                            del bandmeta['band_names']
 
-                ds_out.FlushCache()
+                        if 'nodata' in bandmeta:
+                            band.SetNoDataValue(self.metadata.band_meta['nodata'][bidx])
+                            del bandmeta['nodata']
+
+                        if bandmeta:
+                            band.SetMetadata(bandmeta)
+
+                        band.FlushCache()
+                        del band
+
+                    ds_out.FlushCache()
+
+                # write ds_out to disk,
+                # -> writes the in-memory array or transforms the linked dataset into the target format
+                gdal.Translate(out_path, ds_out, format=fmt, creationOptions=creationOptions)
+                del ds_src
 
         finally:
+            if 'GDAL_PAM_ENABLED' in os.environ:
+                del os.environ['GDAL_PAM_ENABLED']
+
+            if not os.path.exists(out_path):
+                raise Exception(gdal.GetLastErrorMsg())
+
             del ds_out
 
     def dump(self, out_path):
@@ -1130,7 +1163,7 @@ class GeoArray(object):
                              nodataVal=None, out_prj=None, ignore_rotation=False):
         # handle limits
         if boundsMap:
-            boundsMapPrj = boundsMapPrj if boundsMapPrj else self.prj
+            boundsMapPrj = boundsMapPrj or self.prj
             image2plot, gt, prj = self.get_mapPos(boundsMap, boundsMapPrj, band2get=band,
                                                   fillVal=nodataVal if nodataVal is not None else self.nodata)
         else:
@@ -1146,9 +1179,11 @@ class GeoArray(object):
         out_nodata = in_nodata if in_nodata is not None else -9999
         if not np.can_cast(out_nodata, image2plot.dtype):
             image2plot = image2plot.astype(np.int32)
+        if image2plot.dtype == bool:
+            image2plot = image2plot.astype(int)
 
         # rotated images always have to be resampled for plotting
-        if not ignore_rotation and (gt[2] or gt[4]):
+        if not ignore_rotation and self.is_rotated:
             out_prj = out_prj or self.projection
 
         if res_factor != 1. and image2plot.shape[0] * image2plot.shape[1] > 1e6:  # shape > 1000*1000
@@ -1297,7 +1332,7 @@ class GeoArray(object):
                 ax = plt.gca()
 
             rows, cols = image2plot.shape[:2]
-            im = ax.imshow(image2plot.astype(int),
+            im = ax.imshow(image2plot,
                            palette,
                            interpolation=interpolation,
                            extent=(0, cols, rows, 0),
@@ -1342,22 +1377,30 @@ class GeoArray(object):
         # get image to plot
         # (reproject to LonLat as workaround in case self.epsg is None because cartopy relies on an existing EPSG code)
         nodataVal = nodataVal if nodataVal is not None else self.nodata
-        gA2plot = GeoArray(*self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap,
-                                                      boundsMapPrj=boundsMapPrj, res_factor=res_factor,
+        gA2plot = GeoArray(*self._get_plottable_image(xlim, ylim, band,
+                                                      boundsMap=boundsMap,
+                                                      boundsMapPrj=boundsMapPrj,
+                                                      res_factor=res_factor,
                                                       nodataVal=nodataVal,
-                                                      out_prj=None if self.epsg is not None else 4326),
+                                                      # FIXME EPSG:4326 fails for extraterrestrial data
+                                                      out_prj=self.epsg or 4326
+                                                      ),
                            nodata=nodataVal)
         image2plot = gA2plot[:]
 
         # create map
         def get_cartopy_crs_from_epsg(epsg_code):
-            try:
-                return ccrs_from_epsg(epsg_code)
-            except ValueError:
-                if epsg_code == 4326:
-                    return PlateCarree()
-                else:
-                    raise NotImplementedError('The show_map() method currently does not support the given projection.')
+            if epsg_code:
+                try:
+                    return ccrs_from_epsg(epsg_code)
+                except ValueError:
+                    if epsg_code == 4326:
+                        return PlateCarree()
+                    else:
+                        raise NotImplementedError('The show_map() method currently does not support the given '
+                                                  'projection.')
+            else:
+                raise ValueError(f'Expected a valid EPSG code. Got {epsg_code}.')
 
         crs_in = get_cartopy_crs_from_epsg(gA2plot.epsg)
         crs_out = get_cartopy_crs_from_epsg(out_epsg if out_epsg is not None else gA2plot.epsg)
@@ -1558,14 +1601,14 @@ class GeoArray(object):
 
         :param poly: instance of shapely.geometry.Polygon
         """
-        self.arr, self.gt, self.projection = self.get_mapPos(mapBounds=poly.bounds, mapBounds_prj=self.prj)
+        self.arr, self.gt, self.projection = self.get_mapPos(mapBounds=poly.bounds)
         self.mask_nodata.arr, self.mask_nodata.gt, self.mask_nodata.projection = \
             self.mask_nodata.get_mapPos(mapBounds=poly.bounds, mapBounds_prj=self.prj)
         assert self.shape == self.mask_nodata.shape
 
         if self._mask_baddata is not None:
             self.mask_baddata.arr, self.mask_baddata.gt, self.mask_baddata.projection = \
-                self.mask_baddata.get_mapPos(mapBounds=poly.bounds, mapBounds_prj=self.prj)
+                self.mask_baddata.get_mapPos(mapBounds=poly.bounds)
             assert self.shape == self.mask_baddata.shape
 
         # update footprint polygon
@@ -1605,8 +1648,8 @@ class GeoArray(object):
 
         return GeneratorLen(out_gen, length=len(bounds_alltiles))
 
-    def get_mapPos(self, mapBounds, mapBounds_prj, band2get=None, out_prj=None, arr_gt=None, arr_prj=None, fillVal=None,
-                   rspAlg='near', progress=None, v=False):  # TODO implement slice for indexing bands
+    def get_mapPos(self, mapBounds, mapBounds_prj=None, band2get=None, out_prj=None, arr_gt=None, arr_prj=None,
+                   fillVal=None, rspAlg='near', progress=None, v=False):  # TODO implement slice for indexing bands
         # type: (tuple, str, int, str, tuple, str, int, str, bool, bool) -> (np.ndarray, tuple, str)
         """Return the array data of GeoArray at a given geographic position.
 
@@ -1626,6 +1669,7 @@ class GeoArray(object):
         :param v:               verbose mode (not related to GeoArray.v; must be explicitly set)
         :return:
         """
+        mapBounds_prj = mapBounds_prj if mapBounds_prj is not None else self.prj
         arr_gt = arr_gt or self.geotransform
         arr_prj = arr_prj or self.projection
         out_prj = out_prj or arr_prj
@@ -1728,10 +1772,10 @@ class GeoArray(object):
 
         # apply zslice to bandnames and metadata
         if zslicing:
-            bNs_out = list(np.array(list(self.bandnames))[zslice])
+            bNs_out = list(np.array(list(self._bandnames))[zslice]) if self._bandnames else None
             _meta_out = self.metadata.get_subset(bands2extract=zslice)
         else:
-            bNs_out = list(self.bandnames)
+            bNs_out = list(self._bandnames) if self._bandnames else None
             _meta_out = self.meta
 
         sub_gA.gt = sub_gt
@@ -1904,7 +1948,7 @@ class GeoArray(object):
 
 def get_GeoArray_from_GDAL_ds(ds):
     # TODO implement as class method of GeoArray
-    arr = gdalnumeric.DatasetReadAsArray(ds)
+    arr = gdal_array.DatasetReadAsArray(ds)
     if len(arr.shape) == 3:
         arr = np.swapaxes(np.swapaxes(arr, 0, 2), 0, 1)
     return GeoArray(arr, ds.GetGeoTransform(), ds.GetProjection())
